from __future__ import absolute_import, unicode_literals

__all__ = ['Pregel']

from .api import APIWrapper
from .exceptions import (
    PregelJobGetError,
    PregelJobCreateError,
    PregelJobDeleteError
)
from .request import Request


class Pregel(APIWrapper):
    """Pregel API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    """

    def __init__(self, connection, executor):
        super(Pregel, self).__init__(connection, executor)

    def __repr__(self):
        return '<Pregel in {}>'.format(self._conn.db_name)

    async def job(self, job_id):
        """Return the details of a Pregel job.

        :param job_id: Pregel job ID.
        :type job_id: int
        :return: Details of the Pregel job.
        :rtype: dict
        :raise arango.exceptions.PregelJobGetError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/control_pregel/{}'.format(job_id)
        )

        def response_handler(resp):
            if not resp.is_success:
                raise PregelJobGetError(resp, request)
            if 'receivedCount' in resp.body:
                resp.body['received_count'] = resp.body.pop('receivedCount')
            if 'sendCount' in resp.body:
                resp.body['send_count'] = resp.body.pop('sendCount')
            if 'totalRuntime' in resp.body:
                resp.body['total_runtime'] = resp.body.pop('totalRuntime')
            if 'edgeCount' in resp.body:  # pragma: no cover
                resp.body['edge_count'] = resp.body.pop('edgeCount')
            if 'vertexCount' in resp.body:  # pragma: no cover
                resp.body['vertex_count'] = resp.body.pop('vertexCount')
            return resp.body

        return await self._execute(request, response_handler)

    async def create_job(
            self,
            graph,
            algorithm,
            store=True,
            max_gss=None,
            thread_count=None,
            async_mode=None,
            result_field=None,
            algorithm_params=None):
        """Start a new Pregel job.

        :param graph: Graph name.
        :type graph: str | unicode
        :param algorithm: Algorithm (e.g. "pagerank").
        :type algorithm: str | unicode
        :param store: If set to True, Pregel engine writes results back to the
            database. If set to False, results can be queried via AQL.
        :type store: bool
        :param max_gss: Max number of global iterations for the algorithm.
        :type max_gss: int
        :param thread_count: Number of parallel threads to use per worker.
            This does not influence the number of threads used to load or store
            data from the database (it depends on the number of shards).
        :type thread_count: int
        :param async_mode: If set to True, algorithms which support async mode
            run without synchronized global iterations. This might lead to
            performance increase if there are load imbalances.
        :type async_mode: bool
        :param result_field: If specified, most algorithms will write their
            results into this field.
        :type result_field: str | unicode
        :param algorithm_params: Additional algorithm parameters.
        :type algorithm_params: dict
        :return: Pregel job ID.
        :rtype: int
        :raise arango.exceptions.PregelJobCreateError: If create fails.
        """
        data = {'algorithm': algorithm, 'graphName': graph}

        if algorithm_params is None:
            algorithm_params = {}

        if store is not None:
            algorithm_params['store'] = store
        if max_gss is not None:
            algorithm_params['maxGSS'] = max_gss
        if thread_count is not None:
            algorithm_params['parallelism'] = thread_count
        if async_mode is not None:
            algorithm_params['async'] = async_mode
        if result_field is not None:
            algorithm_params['resultField'] = result_field
        if algorithm_params:
            data['params'] = algorithm_params

        request = Request(
            method='post',
            endpoint='/_api/control_pregel',
            data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body
            raise PregelJobCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_job(self, job_id):
        """Delete a Pregel job.

        :param job_id: Pregel job ID.
        :type job_id: int
        :return: True if Pregel job was deleted successfully.
        :rtype: bool
        :raise arango.exceptions.PregelJobDeleteError: If delete fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/control_pregel/{}'.format(job_id)
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            raise PregelJobDeleteError(resp, request)

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

import json

import mock
import pytest
from six import string_types

from aioarangodb.database import BatchDatabase
from aioarangodb.exceptions import (
    DocumentInsertError,
    BatchExecuteError,
    BatchJobResultError,
    BatchStateError
)
from aioarangodb.job import BatchJob
from aioarangodb.tests.helpers import extract, clean_doc
pytestmark = pytest.mark.asyncio

async def test_batch_wrapper_attributes(db, col, username):
    batch_db = db.begin_batch_execution()
    assert isinstance(batch_db, BatchDatabase)
    assert batch_db.username == username
    assert batch_db.context == 'batch'
    assert batch_db.db_name == db.name
    assert batch_db.name == db.name
    assert repr(batch_db) == '<BatchDatabase {}>'.format(db.name)

    batch_col = batch_db.collection(col.name)
    assert batch_col.username == username
    assert batch_col.context == 'batch'
    assert batch_col.db_name == db.name
    assert batch_col.name == col.name

    batch_aql = batch_db.aql
    assert batch_aql.username == username
    assert batch_aql.context == 'batch'
    assert batch_aql.db_name == db.name

    job = await batch_aql.execute('INVALID QUERY')
    assert isinstance(job, BatchJob)
    assert isinstance(job.id, string_types)
    assert repr(job) == '<BatchJob {}>'.format(job.id)


async def test_batch_execute_without_result(db, col, docs):
    async with db.begin_batch_execution(return_result=False) as batch_db:
        batch_col = batch_db.collection(col.name)
        # Ensure that no jobs are returned
        assert await batch_col.insert(docs[0]) is None
        assert await batch_col.delete(docs[0]) is None
        assert await batch_col.insert(docs[1]) is None
        assert await batch_col.delete(docs[1]) is None
        assert await batch_col.insert(docs[2]) is None
        assert await batch_col.get(docs[2]) is None
        assert batch_db.queued_jobs() is None

    # Ensure that the operations went through
    assert batch_db.queued_jobs() is None
    assert await extract('_key', await col.all()) == [docs[2]['_key']]


async def test_batch_execute_with_result(db, col, docs):
    async with db.begin_batch_execution(return_result=True) as batch_db:
        batch_col = batch_db.collection(col.name)
        job1 = await batch_col.insert(docs[0])
        job2 = await batch_col.insert(docs[1])
        job3 = await batch_col.insert(docs[1])  # duplicate
        jobs = batch_db.queued_jobs()
        assert jobs == [job1, job2, job3]
        assert all(job.status() == 'pending' for job in jobs)

    assert batch_db.queued_jobs() == [job1, job2, job3]
    assert all(job.status() == 'done' for job in batch_db.queued_jobs())
    assert await extract('_key', await col.all()) == await extract('_key', docs[:2])

    # Test successful results
    assert job1.result()['_key'] == docs[0]['_key']
    assert job2.result()['_key'] == docs[1]['_key']

    # Test insert error result
    with pytest.raises(DocumentInsertError) as err:
        await job3.result()
    assert err.value.error_code == 1210


async def test_batch_empty_commit(db):
    batch_db = db.begin_batch_execution(return_result=False)
    assert await batch_db.commit() is None

    batch_db = db.begin_batch_execution(return_result=True)
    assert await batch_db.commit() == []


async def test_batch_double_commit(db, col, docs):
    batch_db = db.begin_batch_execution()
    job = await batch_db.collection(col.name).insert(docs[0])

    # Test first commit
    assert await batch_db.commit() == [job]
    assert job.status() == 'done'
    assert await col.count() == 1
    assert await clean_doc(await col.random()) == docs[0]

    # Test second commit which should fail
    with pytest.raises(BatchStateError) as err:
        await batch_db.commit()
    assert 'already committed' in str(err.value)
    assert await col.count() == 1
    assert await clean_doc(await col.random()) == docs[0]


async def test_batch_action_after_commit(db, col):
    async with db.begin_batch_execution() as batch_db:
        await batch_db.collection(col.name).insert({})

    # Test insert after the batch has been committed
    with pytest.raises(BatchStateError) as err:
        await batch_db.collection(col.name).insert({})
    assert 'already committed' in str(err.value)
    assert await col.count() == 1


async def test_batch_execute_error(bad_db, col, docs):
    batch_db = bad_db.begin_batch_execution(return_result=True)
    job = await batch_db.collection(col.name).insert_many(docs)

    # Test batch execute with bad database
    with pytest.raises(BatchExecuteError) as err:
        await batch_db.commit()
    assert err.value.error_code in {11, 1228}
    assert await col.count() == 0
    assert job.status() == 'pending'


async def test_batch_job_result_not_ready(db, col, docs):
    batch_db = db.begin_batch_execution(return_result=True)
    job = await batch_db.collection(col.name).insert_many(docs)

    # Test get job result before commit
    with pytest.raises(BatchJobResultError) as err:
        await job.result()
    assert str(err.value) == 'result not available yet'

    # Test commit to make sure it still works after the errors
    assert await batch_db.commit() == [job]
    assert len(job.result()) == len(docs)
    assert await extract('_key', await col.all()) == await extract('_key', docs)


class AsyncMock(mock.MagicMock):
    async def __call__(self, *args, **kwargs):
        return super(AsyncMock, self).__call__(*args, **kwargs)

async def test_batch_bad_state(db, col, docs):
    batch_db = db.begin_batch_execution()
    batch_col = batch_db.collection(col.name)
    await batch_col.insert(docs[0])
    await batch_col.insert(docs[1])
    await batch_col.insert(docs[2])

    # Monkey patch the connection object
    mock_resp = mock.MagicMock()
    mock_resp.is_success = True
    mock_resp.raw_body = ''
    mock_send_request = AsyncMock()
    mock_send_request.return_value = mock_resp
    mock_connection = mock.MagicMock()
    mock_connection.send_request = mock_send_request
    mock_connection.serialize = json.dumps
    mock_connection.deserialize = json.loads
    batch_db._executor._conn = mock_connection

    # Test commit with invalid batch state
    with pytest.raises(BatchStateError) as err:
        await batch_db.commit()
    assert 'expecting 3 parts in batch response but got 0' in str(err.value)



from __future__ import absolute_import, unicode_literals

from six import string_types
import pytest

from aioarangodb.exceptions import (
    TaskCreateError,
    TaskDeleteError,
    TaskGetError,
    TaskListError
)
from aioarangodb.tests.helpers import (
    assert_raises,
    extract,
    generate_task_id,
    generate_task_name,
)
pytestmark = pytest.mark.asyncio


async def test_task_management(sys_db, db, bad_db):
    test_command = 'require("@arangodb").print(params);'

    # Test create task with random ID
    task_name = generate_task_name()
    new_task = await db.create_task(
        name=task_name,
        command=test_command,
        params={'foo': 1, 'bar': 2},
        offset=1,
    )
    assert new_task['name'] == task_name
    assert 'print(params)' in new_task['command']
    assert new_task['type'] == 'timed'
    assert new_task['database'] == db.name
    assert isinstance(new_task['created'], float)
    assert isinstance(new_task['id'], string_types)

    # Test get existing task
    assert await db.task(new_task['id']) == new_task

    # Test create task with specific ID
    task_name = generate_task_name()
    task_id = generate_task_id()
    new_task = await db.create_task(
        name=task_name,
        command=test_command,
        params={'foo': 1, 'bar': 2},
        offset=1,
        period=10,
        task_id=task_id
    )
    assert new_task['name'] == task_name
    assert new_task['id'] == task_id
    assert 'print(params)' in new_task['command']
    assert new_task['type'] == 'periodic'
    assert new_task['database'] == db.name
    assert isinstance(new_task['created'], float)
    assert await db.task(new_task['id']) == new_task

    # Test create duplicate task
    with assert_raises(TaskCreateError) as err:
        await db.create_task(
            name=task_name,
            command=test_command,
            params={'foo': 1, 'bar': 2},
            task_id=task_id
        )
    assert err.value.error_code == 1851

    # Test list tasks
    for task in await sys_db.tasks():
        assert task['type'] in {'periodic', 'timed'}
        assert isinstance(task['id'], string_types)
        assert isinstance(task['name'], string_types)
        assert isinstance(task['created'], float)
        assert isinstance(task['command'], string_types)

    # Test list tasks with bad database
    with assert_raises(TaskListError) as err:
        await bad_db.tasks()
    assert err.value.error_code in {11, 1228}

    # Test get missing task
    with assert_raises(TaskGetError) as err:
        await db.task(generate_task_id())
    assert err.value.error_code == 1852

    # Test delete existing task
    assert task_id in await extract('id', await db.tasks())
    assert await db.delete_task(task_id) is True
    assert task_id not in await extract('id', await db.tasks())
    with assert_raises(TaskGetError) as err:
        await db.task(task_id)
    assert err.value.error_code == 1852

    # Test delete missing task
    with assert_raises(TaskDeleteError) as err:
        await db.delete_task(generate_task_id(), ignore_missing=False)
    assert err.value.error_code == 1852
    assert await db.delete_task(task_id, ignore_missing=True) is False



from __future__ import absolute_import, unicode_literals

__all__ = ['Replication']

from .api import APIWrapper
from .exceptions import (
    ReplicationApplierConfigError,
    ReplicationApplierConfigSetError,
    ReplicationApplierStartError,
    ReplicationApplierStateError,
    ReplicationApplierStopError,
    ReplicationClusterInventoryError,
    ReplicationDumpBatchCreateError,
    ReplicationDumpBatchDeleteError,
    ReplicationDumpBatchExtendError,
    ReplicationDumpError,
    ReplicationInventoryError,
    ReplicationLoggerFirstTickError,
    ReplicationLoggerStateError,
    ReplicationMakeSlaveError,
    ReplicationServerIDError,
    ReplicationSyncError
)
from .formatter import (
    format_replication_applier_config,
    format_replication_applier_state,
    format_replication_header,
    format_replication_inventory,
    format_replication_logger_state,
    format_replication_sync
)
from .request import Request


class Replication(APIWrapper):

    def __init__(self, connection, executor):
        super(Replication, self).__init__(connection, executor)

    async def inventory(self, batch_id, include_system=None, all_databases=None):
        """Return an overview of collections and indexes.

        :param batch_id: Batch ID. For RocksDB engine only.
        :type batch_id: str | unicode
        :param include_system: Include system collections in the result.
            Default value is True.
        :type include_system: bool
        :param all_databases: Include all databases. Only works on "_system"
            database. Default value is False.
        :type all_databases: bool
        :return: Overview of collections and indexes.
        :rtype: dict
        :raise arango.exceptions.ReplicationInventoryError: If retrieval fails.
        """
        params = {'batchId': batch_id}
        if include_system is not None:
            params['includeSystem'] = include_system
        if all_databases is not None:
            params['global'] = all_databases

        request = Request(
            method='get',
            endpoint='/_api/replication/inventory',
            params=params
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_inventory(resp.body)
            raise ReplicationInventoryError(resp, request)

        return await self._execute(request, response_handler)

    async def create_dump_batch(self, ttl=None):
        """Create a new dump batch.

        :param ttl: Time-to-live for the new batch in seconds.
        :type ttl: int
        :return: ID of the batch.
        :rtype: dict
        :raise arango.exceptions.ReplicationDumpBatchCreateError: If create
            fails.
        """
        request = Request(
            method='post',
            endpoint='/_api/replication/batch',
            data={'ttl': ttl}
        )

        def response_handler(resp):
            if resp.is_success:
                return {
                    'id': resp.body['id'],
                    'last_tick': resp.body['lastTick']
                }
            raise ReplicationDumpBatchCreateError(resp, request)

        return await self._execute(request, response_handler)

    async def delete_dump_batch(self, batch_id):
        """Delete a dump batch.

        :param batch_id: Dump batch ID.
        :type batch_id: str | unicode
        :return: True if deletion was successful.
        :rtype: bool
        :raise arango.exceptions.ReplicationDumpBatchDeleteError: If delete
            fails.
        """
        request = Request(
            method='delete',
            endpoint='/_api/replication/batch/{}'.format(batch_id),
            deserialize=False
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            raise ReplicationDumpBatchDeleteError(resp, request)

        return await self._execute(request, response_handler)

    async def extend_dump_batch(self, batch_id, ttl):
        """Extend a dump batch.

        :param batch_id: Dump batch ID.
        :type batch_id: str | unicode
        :param ttl: Time-to-live for the new batch in seconds.
        :type ttl: int
        :return: True if operation was successful.
        :rtype: bool
        :raise arango.exceptions.ReplicationDumpBatchExtendError: If operation
            fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/replication/batch/{}'.format(batch_id),
            data={'ttl': ttl},
            deserialize=False
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            raise ReplicationDumpBatchExtendError(resp, request)

        return await self._execute(request, response_handler)

    async def dump(
            self,
            collection,
            batch_id=None,
            lower=None,
            upper=None,
            chunk_size=None,
            include_system=None,
            ticks=None,
            flush=None,
            deserialize=False):
        """Return the events data of one collection.

        :param collection: Name or ID of the collection to dump.
        :type collection: str | unicode
        :param chunk_size: Size of the result in bytes. This value is honored
            approximately only.
        :type chunk_size: int
        :param batch_id: Batch ID. For RocksDB engine only.
        :type batch_id: str | unicode
        :param lower: Lower bound tick value for results. For MMFiles only.
        :type lower: str | unicode
        :param upper: Upper bound tick value for results. For MMFiles only.
        :type upper: str | unicode
        :param include_system: Include system collections in the result. For
            MMFiles only. Default value is True.
        :type include_system: bool
        :param ticks: Whether to include tick values in the dump. For MMFiles
            only. Default value is True.
        :type ticks: bool
        :param flush: Whether to flush the WAL before dumping. Default value is
            True.
        :type flush: bool
        :param deserialize: Deserialize the response content. Default is False.
        :type deserialize: bool
        :return: Collection events data.
        :rtype: str | unicode | [dict]
        :raise arango.exceptions.ReplicationDumpError: If retrieval fails.
        """
        params = {'collection': collection}

        if chunk_size is not None:
            params['chunkSize'] = chunk_size
        if batch_id is not None:
            params['batchId'] = batch_id
        if lower is not None:
            params['from'] = lower
        if upper is not None:
            params['to'] = upper
        if include_system is not None:
            params['includeSystem'] = include_system
        if ticks is not None:
            params['ticks'] = ticks
        if flush is not None:
            params['flush '] = flush

        request = Request(
            method='get',
            endpoint='/_api/replication/dump',
            params=params,
            deserialize=False
        )

        def response_handler(resp):
            if resp.is_success:
                result = format_replication_header(resp.headers)
                result['content'] = [
                    self._conn.deserialize(line) for
                    line in resp.body.split('\n') if line
                ] if deserialize else resp.body
                return result

            raise ReplicationDumpError(resp, request)

        return await self._execute(request, response_handler)

    async def synchronize(
            self,
            endpoint,
            database=None,
            username=None,
            password=None,
            include_system=None,
            incremental=None,
            restrict_type=None,
            restrict_collections=None,
            initial_sync_wait_time=None):  # pragma: no cover
        """Synchronize data from a remote endpoint.

        :param endpoint: Master endpoint (e.g. "tcp://192.168.173.13:8529").
        :type endpoint: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param username: Username.
        :type username: str | unicode
        :param password: Password.
        :type password: str | unicode
        :param include_system: Whether to include system collection operations.
        :type include_system: bool
        :param incremental: If set to True, then an incremental synchronization
            method is used for synchronizing data in collections. This
            method is useful when collections already exist locally, and only
            the remaining differences need to be transferred from the remote
            endpoint. In this case, the incremental synchronization can be
            faster than a full synchronization. Default value is False, meaning
            complete data is transferred.
        :type incremental: bool
        :param restrict_type: Optional string value for collection filtering.
            Allowed values are "include" or "exclude".
        :type restrict_type: str | unicode
        :param restrict_collections: Optional list of collections for use with
            argument **restrict_type**. If **restrict_type** set to "include",
            only the specified collections are synchronised. Otherwise, all but
            the specified ones are synchronized.
        :type restrict_collections: [str | unicode]
        :param initial_sync_wait_time: Maximum wait time in seconds that the
            initial synchronization will wait for a response from master when
            fetching collection data. This can be used to control after what
            time the initial synchronization will give up waiting for response
            and fail. Value is ignored if set to 0.
        :type initial_sync_wait_time: int
        :return: Collections transferred and last log tick.
        :rtype: dict
        :raise arango.exceptions.ReplicationSyncError: If sync fails.
        """
        data = {'endpoint': endpoint}

        if database is not None:
            data['database'] = database
        if username is not None:
            data['username'] = username
        if password is not None:
            data['password'] = password
        if include_system is not None:
            data['includeSystem'] = include_system
        if incremental is not None:
            data['incremental'] = incremental
        if restrict_type is not None:
            data['restrictType'] = restrict_type
        if restrict_collections is not None:
            data['restrictCollections'] = restrict_collections
        if initial_sync_wait_time is not None:
            data['initialSyncMaxWaitTime'] = initial_sync_wait_time

        request = Request(
            method='put',
            endpoint='/_api/replication/sync',
            data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_sync(resp.body)
            raise ReplicationSyncError(resp, request)

        return await self._execute(request, response_handler)

    async def cluster_inventory(self, include_system=None):
        """Return an overview of collections and indexes in a cluster.

        :param include_system: Include system collections in the result.
            Default value is True.
        :type include_system: bool
        :return: Overview of collections and indexes on the cluster.
        :rtype: dict
        :raise arango.exceptions.ReplicationClusterInventoryError: If retrieval
            fails.
        """
        params = {}
        if include_system is not None:
            params['includeSystem'] = include_system

        request = Request(
            method='get',
            endpoint='/_api/replication/clusterInventory',
            params=params
        )

        def response_handler(resp):
            if resp.is_success:  # pragma: no cover
                return format_replication_inventory(resp.body)
            raise ReplicationClusterInventoryError(resp, request)

        return await self._execute(request, response_handler)

    async def logger_state(self):
        """Return the state of the replication logger.

        :return: Logger state.
        :rtype: dict
        :raise arango.exceptions.ReplicationLoggerStateError: If retrieval
            fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/replication/logger-state',
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_logger_state(resp.body)
            raise ReplicationLoggerStateError(resp, request)

        return await self._execute(request, response_handler)

    async def logger_first_tick(self):
        """Return the first available tick value from the server.

        :return: First tick value.
        :rtype: str | unicode
        :raise arango.exceptions.ReplicationLoggerFirstTickError: If retrieval
            fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/replication/logger-first-tick',
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body['firstTick']
            raise ReplicationLoggerFirstTickError(resp, request)

        return await self._execute(request, response_handler)

    async def applier_config(self):
        """Return the configuration of the replication applier.

        :return: Configuration of the replication applier.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierConfigError: If retrieval
            fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/replication/applier-config',
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_config(resp.body)
            raise ReplicationApplierConfigError(resp, request)

        return await self._execute(request, response_handler)

    async def set_applier_config(
            self,
            endpoint,
            database=None,
            username=None,
            password=None,
            max_connect_retries=None,
            connect_timeout=None,
            request_timeout=None,
            chunk_size=None,
            auto_start=None,
            adaptive_polling=None,
            include_system=None,
            auto_resync=None,
            auto_resync_retries=None,
            initial_sync_max_wait_time=None,
            connection_retry_wait_time=None,
            idle_min_wait_time=None,
            idle_max_wait_time=None,
            require_from_present=None,
            verbose=None,
            restrict_type=None,
            restrict_collections=None):
        """Set configuration values of the replication applier.

        :param endpoint: Server endpoint (e.g. "tcp://192.168.173.13:8529").
        :type endpoint: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param username: Username.
        :type username: str | unicode
        :param password: Password.
        :type password: str | unicode
        :param max_connect_retries: Maximum number of connection attempts the
            applier makes in a row before stopping itself.
        :type max_connect_retries: int
        :param connect_timeout: Timeout in seconds when attempting to connect
            to the endpoint. This value is used for each connection attempt.
        :type connect_timeout: int
        :param request_timeout: Timeout in seconds for individual requests to
            the endpoint.
        :type request_timeout: int
        :param chunk_size: Requested maximum size in bytes for log transfer
            packets when the endpoint is contacted.
        :type chunk_size: int
        :param auto_start: Whether to auto-start the replication applier on
            (next and following) server starts.
        :type auto_start: bool
        :param adaptive_polling: If set to True, replication applier sleeps
            for an increasingly long period in case the logger server at the
            endpoint has no replication events to apply. Using adaptive polling
            reduces the amount of work done by both the applier and the logger
            server when there are infrequent changes. The downside is that it
            might take longer for the replication applier to detect new events.
        :type adaptive_polling: bool
        :param include_system: Whether system collection operations are
            applied.
        :type include_system: bool
        :param auto_resync: Whether the slave should perform a full automatic
            resynchronization with the master in case the master cannot serve
            log data requested by the slave, or when the replication is started
            and no tick value can be found.
        :type auto_resync: bool
        :param auto_resync_retries: Max number of resynchronization retries.
            Setting this to 0 disables it.
        :type auto_resync_retries: int
        :param initial_sync_max_wait_time: Max wait time in seconds the initial
            synchronization waits for master on collection data. This value
            is relevant even for continuous replication when **auto_resync** is
            set to True because this may re-start the initial synchronization
            when master cannot provide log data slave requires. This value is
            ignored if set to 0.
        :type initial_sync_max_wait_time: int
        :param connection_retry_wait_time: Time in seconds the applier idles
            before trying to connect to master in case of connection problems.
            This value is ignored if set to 0.
        :type connection_retry_wait_time: int
        :param idle_min_wait_time: Minimum wait time in seconds the applier
            idles before fetching more log data from the master in case the
            master has already sent all its log data. This wait time can be
            used to control the frequency with which the replication applier
            sends HTTP log fetch requests to the master in case there is no
            write activity on the master. This value is ignored if set to 0.
        :type idle_min_wait_time: int
        :param idle_max_wait_time: Maximum wait time in seconds the applier
            idles before fetching more log data from the master in case the
            master has already sent all its log data. This wait time can be
            used to control the maximum frequency with which the replication
            applier sends HTTP log fetch requests to the master in case there
            is no write activity on the master. Applies only when argument
            **adaptive_polling** is set to True. This value is ignored if set
            to 0.
        :type idle_max_wait_time: int
        :param require_from_present: If set to True, replication applier checks
            at start whether the start tick from which it starts or resumes
            replication is still present on the master. If not, then there
            would be data loss. If set to True, the replication applier aborts
            with an appropriate error message. If set to False, the applier
            still starts and ignores the data loss.
        :type require_from_present: bool
        :param verbose: If set to True, a log line is emitted for all
            operations performed by the replication applier. This should be
            used for debugging replication problems only.
        :type verbose: bool
        :param restrict_type: Optional string value for collection filtering.
            Allowed values are "include" or "exclude".
        :type restrict_type: str | unicode
        :param restrict_collections: Optional list of collections for use with
            argument **restrict_type**. If **restrict_type** set to "include",
            only the specified collections are included. Otherwise, only the
            specified collections are excluded.
        :type restrict_collections: [str | unicode]
        :return: Updated configuration.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierConfigSetError: If update
            fails.
        """
        data = {'endpoint': endpoint}

        if database is not None:
            data['database'] = database
        if username is not None:
            data['username'] = username
        if password is not None:
            data['password'] = password
        if max_connect_retries is not None:
            data['maxConnectRetries'] = max_connect_retries
        if connect_timeout is not None:
            data['connectTimeout'] = connect_timeout
        if request_timeout is not None:
            data['requestTimeout'] = request_timeout
        if chunk_size is not None:
            data['chunkSize'] = chunk_size
        if auto_start is not None:
            data['autoStart'] = auto_start
        if adaptive_polling is not None:
            data['adaptivePolling'] = adaptive_polling
        if include_system is not None:
            data['includeSystem'] = include_system
        if auto_resync is not None:
            data['autoResync'] = auto_resync
        if auto_resync_retries is not None:
            data['autoResyncRetries'] = auto_resync_retries
        if initial_sync_max_wait_time is not None:
            data['initialSyncMaxWaitTime'] = initial_sync_max_wait_time
        if connection_retry_wait_time is not None:
            data['connectionRetryWaitTime'] = connection_retry_wait_time
        if idle_min_wait_time is not None:
            data['idleMinWaitTime'] = idle_min_wait_time
        if idle_max_wait_time is not None:
            data['idleMaxWaitTime'] = idle_max_wait_time
        if require_from_present is not None:
            data['requireFromPresent'] = require_from_present
        if verbose is not None:
            data['verbose'] = verbose
        if restrict_type is not None:
            data['restrictType'] = restrict_type
        if restrict_collections is not None:
            data['restrictCollections'] = restrict_collections

        request = Request(
            method='put',
            endpoint='/_api/replication/applier-config',
            data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_config(resp.body)
            raise ReplicationApplierConfigSetError(resp, request)

        return await self._execute(request, response_handler)

    async def applier_state(self):
        """Return the state of the replication applier

        :return: Applier state and details.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierStateError: If retrieval
            fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/replication/applier-state',
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_state(resp.body)
            raise ReplicationApplierStateError(resp, request)

        return await self._execute(request, response_handler)

    async def start_applier(self, last_tick=None):
        """Start the replication applier.

        :param last_tick: The remote last log tick value from which to start
            applying replication. If not specified, the last saved tick from
            the previous applier run is used. If there is no previous applier
            state saved, the applier starts at the beginning of the logger
            server's log.
        :type last_tick: str | unicode
        :return: Applier state and details.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierStartError: If operation
            fails.
        """
        data = {}
        if last_tick is not None:
            data['from'] = last_tick

        request = Request(
            method='put',
            endpoint='/_api/replication/applier-start',
            params=data
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_state(resp.body)
            raise ReplicationApplierStartError(resp, request)

        return await self._execute(request, response_handler)

    async def stop_applier(self):
        """Stop the replication applier.

        :return: Applier state and details.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierStopError: If operation
            fails.
        """
        request = Request(
            method='put',
            endpoint='/_api/replication/applier-stop',
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_state(resp.body)
            raise ReplicationApplierStopError(resp, request)

        return await self._execute(request, response_handler)

    async def make_slave(
            self,
            endpoint,
            database=None,
            username=None,
            password=None,
            restrict_type=None,
            restrict_collections=None,
            include_system=None,
            max_connect_retries=None,
            connect_timeout=None,
            request_timeout=None,
            chunk_size=None,
            adaptive_polling=None,
            auto_resync=None,
            auto_resync_retries=None,
            initial_sync_max_wait_time=None,
            connection_retry_wait_time=None,
            idle_min_wait_time=None,
            idle_max_wait_time=None,
            require_from_present=None,
            verbose=None):  # pragma: no cover
        """Change the server role to slave.

        :param endpoint: Master endpoint (e.g. "tcp://192.168.173.13:8529").
        :type endpoint: str | unicode
        :param database: Database name.
        :type database: str | unicode
        :param username: Username.
        :type username: str | unicode
        :param password: Password.
        :type password: str | unicode
        :param include_system: Whether system collection operations are
            applied.
        :type include_system: bool
        :param restrict_type: Optional string value for collection filtering.
            Allowed values are "include" or "exclude".
        :type restrict_type: str | unicode
        :param restrict_collections: Optional list of collections for use with
            argument **restrict_type**. If **restrict_type** set to "include",
            only the specified collections are included. Otherwise, only the
            specified collections are excluded.
        :type restrict_collections: [str | unicode]
        :param max_connect_retries: Maximum number of connection attempts the
            applier makes in a row before stopping itself.
        :type max_connect_retries: int
        :param connect_timeout: Timeout in seconds when attempting to connect
            to the endpoint. This value is used for each connection attempt.
        :type connect_timeout: int
        :param request_timeout: Timeout in seconds for individual requests to
            the endpoint.
        :type request_timeout: int
        :param chunk_size: Requested maximum size in bytes for log transfer
            packets when the endpoint is contacted.
        :type chunk_size: int
        :param adaptive_polling: If set to True, replication applier sleeps
            for an increasingly long period in case the logger server at the
            endpoint has no replication events to apply. Using adaptive polling
            reduces the amount of work done by both the applier and the logger
            server when there are infrequent changes. The downside is that it
            might take longer for the replication applier to detect new events.
        :type adaptive_polling: bool
        :param auto_resync: Whether the slave should perform a full automatic
            resynchronization with the master in case the master cannot serve
            log data requested by the slave, or when the replication is started
            and no tick value can be found.
        :type auto_resync: bool
        :param auto_resync_retries: Max number of resynchronization retries.
            Setting this to 0 disables it.
        :type auto_resync_retries: int
        :param initial_sync_max_wait_time: Max wait time in seconds the initial
            synchronization waits for master on collection data. This value
            is relevant even for continuous replication when **auto_resync** is
            set to True because this may restart the initial synchronization
            when master cannot provide log data slave requires. This value is
            ignored if set to 0.
        :type initial_sync_max_wait_time: int
        :param connection_retry_wait_time: Time in seconds the applier idles
            before trying to connect to master in case of connection problems.
            This value is ignored if set to 0.
        :type connection_retry_wait_time: int
        :param idle_min_wait_time: Minimum wait time in seconds the applier
            idles before fetching more log data from the master in case the
            master has already sent all its log data. This wait time can be
            used to control the frequency with which the replication applier
            sends HTTP log fetch requests to the master in case there is no
            write activity on the master. This value is ignored if set to 0.
        :type idle_min_wait_time: int
        :param idle_max_wait_time: Maximum wait time in seconds the applier
            idles before fetching more log data from the master in case the
            master has already sent all its log data. This wait time can be
            used to control the maximum frequency with which the replication
            applier sends HTTP log fetch requests to the master in case there
            is no write activity on the master. Applies only when argument
            **adaptive_polling** is set to True. This value is ignored if set
            to 0.
        :type idle_max_wait_time: int
        :param require_from_present: If set to True, replication applier checks
            at start whether the start tick from which it starts or resumes
            replication is still present on the master. If not, then there
            would be data loss. If set to True, the replication applier aborts
            with an appropriate error message. If set to False, the applier
            still starts and ignores the data loss.
        :type require_from_present: bool
        :param verbose: If set to True, a log line is emitted for all
            operations performed by the replication applier. This should be
            used for debugging replication problems only.
        :type verbose: bool
        :return: Replication details.
        :rtype: dict
        :raise arango.exceptions.ReplicationApplierStopError: If operation
            fails.
        """
        data = {'endpoint': endpoint}

        if database is not None:
            data['database'] = database
        if username is not None:
            data['username'] = username
        if password is not None:
            data['password'] = password
        if restrict_type is not None:
            data['restrictType'] = restrict_type
        if restrict_collections is not None:
            data['restrictCollections'] = restrict_collections
        if include_system is not None:
            data['includeSystem'] = include_system
        if max_connect_retries is not None:
            data['maxConnectRetries'] = max_connect_retries
        if connect_timeout is not None:
            data['connectTimeout'] = connect_timeout
        if request_timeout is not None:
            data['requestTimeout'] = request_timeout
        if chunk_size is not None:
            data['chunkSize'] = chunk_size
        if adaptive_polling is not None:
            data['adaptivePolling'] = adaptive_polling
        if auto_resync is not None:
            data['autoResync'] = auto_resync
        if auto_resync_retries is not None:
            data['autoResyncRetries'] = auto_resync_retries
        if initial_sync_max_wait_time is not None:
            data['initialSyncMaxWaitTime'] = initial_sync_max_wait_time
        if connection_retry_wait_time is not None:
            data['connectionRetryWaitTime'] = connection_retry_wait_time
        if idle_min_wait_time is not None:
            data['idleMinWaitTime'] = idle_min_wait_time
        if idle_max_wait_time is not None:
            data['idleMaxWaitTime'] = idle_max_wait_time
        if require_from_present is not None:
            data['requireFromPresent'] = require_from_present
        if verbose is not None:
            data['verbose'] = verbose

        request = Request(
            method='put',
            endpoint='/_api/replication/make-slave',
            data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return format_replication_applier_state(resp.body)
            raise ReplicationMakeSlaveError(resp, request)

        return await self._execute(request, response_handler)

    async def server_id(self):
        """Return this server's ID.

        :return: Server ID.
        :rtype: str | unicode
        :raise arango.exceptions.ReplicationServerIDError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_api/replication/server-id',
        )

        def response_handler(resp):
            if resp.is_success:
                return resp.body['serverId']
            raise ReplicationServerIDError(resp, request)

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

__all__ = ['HTTPClient', 'DefaultHTTPClient']

from abc import ABCMeta, abstractmethod

import aiohttp

from .response import Response


class HTTPClient(object):  # pragma: no cover
    """Abstract base class for HTTP clients."""

    __metaclass__ = ABCMeta

    @abstractmethod
    def create_session(self, host):
        """Return a new requests session given the host URL.

        This method must be overridden by the user.

        :param host: ArangoDB host URL.
        :type host: str | unicode
        :returns: Requests session object.
        :rtype: requests.Session
        """
        raise NotImplementedError

    @abstractmethod
    async def send_request(
            self,
            session,
            method,
            url,
            headers=None,
            params=None,
            data=None,
            auth=None):
        """Send an HTTP request.

        This method must be overridden by the user.

        :param session: Requests session object.
        :type session: requests.Session
        :param method: HTTP method in lowercase (e.g. "post").
        :type method: str | unicode
        :param url: Request URL.
        :type url: str | unicode
        :param headers: Request headers.
        :type headers: dict
        :param params: URL (query) parameters.
        :type params: dict
        :param data: Request payload.
        :type data: str | unicode | bool | int | list | dict
        :param auth: Username and password.
        :type auth: tuple
        :returns: HTTP response.
        :rtype: arango.response.Response
        """
        raise NotImplementedError


class DefaultHTTPClient(HTTPClient):
    """Default HTTP client implementation."""

    def create_session(self, host):
        """Create and return a new session/connection.

        :param host: ArangoDB host URL.
        :type host: str | unicode
        :returns: requests session object
        :rtype: requests.Session
        """
        return aiohttp.ClientSession()

    async def send_request(
            self,
            session,
            method,
            url,
            params=None,
            data=None,
            headers=None,
            auth=None):
        """Send an HTTP request.

        :param session: Requests session object.
        :type session: requests.Session
        :param method: HTTP method in lowercase (e.g. "post").
        :type method: str | unicode
        :param url: Request URL.
        :type url: str | unicode
        :param headers: Request headers.
        :type headers: dict
        :param params: URL (query) parameters.
        :type params: dict
        :param data: Request payload.
        :type data: str | unicode | bool | int | list | dict
        :param auth: Username and password.
        :type auth: tuple
        :returns: HTTP response.
        :rtype: arango.response.Response
        """
        request = getattr(session, method)
        if auth is not None:
            auth = aiohttp.BasicAuth(auth[0], auth[1])
        response = await request(
            url=url,
            params=params,
            data=data,
            headers=headers,
            auth=auth
        )
        return Response(
            method=method,
            url=url,
            headers=response.headers,
            status_code=response.status,
            status_text=response.reason,
            raw_body=await response.text(),
        )



from __future__ import absolute_import, unicode_literals

import logging
from contextlib import contextmanager

from six import string_types

from .exceptions import DocumentParseError


@contextmanager
def suppress_warning(logger_name):
    """Suppress logger messages.

    :param logger_name: Full name of the logger.
    :type logger_name: str | unicode
    """
    logger = logging.getLogger(logger_name)
    original_log_level = logger.getEffectiveLevel()
    logger.setLevel(logging.CRITICAL)
    yield
    logger.setLevel(original_log_level)


def get_col_name(doc):
    """Return the collection name from input.

    :param doc: Document ID or body with "_id" field.
    :type doc: str | unicode | dict
    :return: Collection name.
    :rtype: [str | unicode]
    :raise arango.exceptions.DocumentParseError: If document ID is missing.
    """
    try:
        doc_id = doc['_id'] if isinstance(doc, dict) else doc
    except KeyError:
        raise DocumentParseError('field "_id" required')
    return doc_id.split('/', 1)[0]


def get_doc_id(doc):
    """Return the document ID from input.

    :param doc: Document ID or body with "_id" field.
    :type doc: str | unicode | dict
    :return: Document ID.
    :rtype: str | unicode
    :raise arango.exceptions.DocumentParseError: If document ID is missing.
    """
    try:
        return doc['_id'] if isinstance(doc, dict) else doc
    except KeyError:
        raise DocumentParseError('field "_id" required')


def is_none_or_int(obj):
    """Check if obj is None or an integer.

    :param obj: Object to check.
    :type obj: object
    :return: True if object is None or an integer.
    :rtype: bool
    """
    return obj is None or (isinstance(obj, int) and obj >= 0)


def is_none_or_str(obj):
    """Check if obj is None or a string.

    :param obj: Object to check.
    :type obj: object
    :return: True if object is None or a string.
    :rtype: bool
    """
    return obj is None or isinstance(obj, string_types)



from __future__ import absolute_import, unicode_literals

__all__ = ['WAL']

from .api import APIWrapper
from .exceptions import (
    WALFlushError,
    WALPropertiesError,
    WALConfigureError,
    WALTransactionListError,
    WALLastTickError,
    WALTickRangesError,
    WALTailError
)
from .formatter import (
    format_replication_header,
    format_tick_values,
    format_wal_properties,
    format_wal_transactions
)
from .request import Request


class WAL(APIWrapper):  # pragma: no cover
    """WAL (Write-Ahead Log) API wrapper.

    :param connection: HTTP connection.
    :type connection: arango.connection.Connection
    :param executor: API executor.
    :type executor: arango.executor.Executor
    """

    def __init__(self, connection, executor):
        super(WAL, self).__init__(connection, executor)

    async def properties(self):
        """Return WAL properties.

        :return: WAL properties.
        :rtype: dict
        :raise arango.exceptions.WALPropertiesError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/wal/properties'
        )

        def response_handler(resp):
            if resp.is_success:
                return format_wal_properties(resp.body)
            raise WALPropertiesError(resp, request)

        return await self._execute(request, response_handler)

    async def configure(
            self,
            oversized_ops=None,
            log_size=None,
            historic_logs=None,
            reserve_logs=None,
            throttle_wait=None,
            throttle_limit=None):
        """Configure WAL properties.

        :param oversized_ops: If set to True, operations bigger than a single
            log file are allowed to be executed and stored.
        :type oversized_ops: bool
        :param log_size: Size of each write-ahead log file in bytes.
        :type log_size: int
        :param historic_logs: Max number of historic log files to keep.
        :type historic_logs: int
        :param reserve_logs: Max number of reserve log files to allocate.
        :type reserve_logs: int
        :param throttle_wait: Wait time before aborting when write-throttled
            in milliseconds.
        :type throttle_wait: int
        :param throttle_limit: Number of pending garbage collector operations
            that, when reached, activates write-throttling. Value of 0 means
            no throttling is triggered.
        :type throttle_limit: int
        :return: New WAL properties.
        :rtype: dict
        :raise arango.exceptions.WALConfigureError: If operation fails.
        """
        data = {}
        if oversized_ops is not None:
            data['allowOversizeEntries'] = oversized_ops
        if log_size is not None:
            data['logfileSize'] = log_size
        if historic_logs is not None:
            data['historicLogfiles'] = historic_logs
        if reserve_logs is not None:
            data['reserveLogfiles'] = reserve_logs
        if throttle_wait is not None:
            data['throttleWait'] = throttle_wait
        if throttle_limit is not None:
            data['throttleWhenPending'] = throttle_limit

        request = Request(
            method='put',
            endpoint='/_admin/wal/properties',
            data=data
        )

        def response_handler(resp):
            if resp.is_success:
                return format_wal_properties(resp.body)
            raise WALConfigureError(resp, request)

        return await self._execute(request, response_handler)

    async def transactions(self):
        """Return details on currently running WAL transactions.

        Fields in the returned details are as follows:

        .. code-block:: none

            "last_collected"    : ID of the last collected log file (at the
                                  start of each running transaction) or None
                                  if no transactions are running.

            "last_sealed"       : ID of the last sealed log file (at the start
                                  of each running transaction) or None if no
                                  transactions are running.

            "count"             : Number of currently running transactions.

        :return: Details on currently running WAL transactions.
        :rtype: dict
        :raise arango.exceptions.WALTransactionListError: If retrieval fails.
        """
        request = Request(
            method='get',
            endpoint='/_admin/wal/transactions'
        )

        def response_handler(resp):
            if resp.is_success:
                return format_wal_transactions(resp.body)
            raise WALTransactionListError(resp, request)

        return await self._execute(request, response_handler)

    async def flush(self, sync=True, garbage_collect=True):
        """Synchronize WAL to disk.

        :param sync: Block until the synchronization is complete.
        :type sync: bool
        :param garbage_collect: Block until flushed data is garbage collected.
        :type garbage_collect: bool
        :return: True if WAL was flushed successfully.
        :rtype: bool
        :raise arango.exceptions.WALFlushError: If flush operation fails.
        """
        request = Request(
            method='put',
            endpoint='/_admin/wal/flush',
            params={
                'waitForSync': sync,
                'waitForCollector': garbage_collect
            }
        )

        def response_handler(resp):
            if resp.is_success:
                return True
            raise WALFlushError(resp, request)

        return await self._execute(request, response_handler)

    async def tick_ranges(self):
        """Return the available ranges of tick values for all WAL files.

        :return: Ranges of tick values.
        :rtype: dict
        :raise arango.exceptions.WALTickRangesError: If retrieval fails.
        """
        request = Request(method='get', endpoint='/_api/wal/range')

        def response_handler(resp):
            if resp.is_success:
                return format_tick_values(resp.body)
            raise WALTickRangesError(resp, request)

        return await self._execute(request, response_handler)

    async def last_tick(self):
        """Return the last available tick value (last successful operation).

        :return: Last tick value in the WAL.
        :rtype: dict
        :raise arango.exceptions.WALLastTickError: If retrieval fails.
        """
        request = Request(method='get', endpoint='/_api/wal/lastTick')

        def response_handler(resp):
            if resp.is_success:
                return format_tick_values(resp.body)
            raise WALLastTickError(resp, request)

        return await self._execute(request, response_handler)

    async def tail(
            self,
            lower=None,
            upper=None,
            last_scanned=None,
            all_databases=None,
            chunk_size=None,
            syncer_id=None,
            server_id=None,
            client_info=None,
            barrier_id=None,
            deserialize=False):
        """Fetch recent WAL operations.

        :param lower: Exclusive lower bound tick value. On successive calls to
            this method you should set this to the value of "last_included"
            from previous call (unless the value was 0).
        :type lower: str | unicode
        :param upper: Inclusive upper bound tick value for results.
        :type upper: str | unicode
        :param last_scanned: On successive calls to this method you should
            set this to the value of "last_scanned" from previous call (or 0
            on first try). This allows the rocksdb engine to break up large
            transactions over multiple responses.
        :type last_scanned: str | unicode
        :param all_databases: Whether operations for all databases should be
            included. When set to False only the operations for the current
            database are included. The value True is only valid on "_system"
            database. The default is False.
        :type all_databases: bool
        :param chunk_size: Approximate maximum size of the returned result.
        :type chunk_size: int
        :param syncer_id: ID of the client used to tail results. The server
            will use this to keep operations until the client has fetched them.
            Must be a positive integer. Note this or **server_id** is required
            to have a chance at fetching reading all operations with the
            rocksdb storage engine.
        :type syncer_id: int
        :param server_id: ID of the client machine. If unset, the server will
            use this to keep operations until the client has fetched them. Must
            be a positive integer. Note this or **syncer_id** is required to
            have a chance at fetching reading all operations with the rocksdb
            storage engine.
        :type server_id: int
        :param client_info: Short description of the client, used for
            informative purposes only.
        :type client_info: str | unicode
        :param barrier_id: ID of barrier used to keep WAL entries around. Only
            required for the MMFiles storage engine.
        :type barrier_id: int
        :param deserialize: Deserialize the response content. Default is False.
        :type deserialize: bool
        :return: If **deserialize** is set to False, content is returned raw as
            a string. If **deserialize** is set to True, it is deserialized and
            returned as a list of dictionaries.
        :rtype: str | [dict]
        :raise arango.exceptions.WALTailError: If tail operation fails.
        """
        params = {}
        if lower is not None:
            params['from'] = lower
        if lower is not None and upper is not None:
            params['to'] = upper
        if lower is not None and last_scanned is not None:
            params['lastScanned'] = last_scanned
        if lower is not None and all_databases is not None:
            params['global'] = all_databases
        if lower is not None and chunk_size is not None:
            params['chunkSize'] = chunk_size
        if lower is not None and syncer_id is not None:
            params['syncerId'] = syncer_id
        if lower is not None and server_id is not None:
            params['serverId'] = server_id
        if lower is not None and client_info is not None:
            params['clientInfo'] = client_info
        if lower is not None and barrier_id is not None:
            params['barrierId'] = barrier_id

        request = Request(
            method='get',
            endpoint='/_api/wal/tail',
            params=params,
            deserialize=False
        )

        def response_handler(resp):
            if resp.is_success:
                result = format_replication_header(resp.headers)
                result['content'] = [
                    self._conn.deserialize(line) for
                    line in resp.body.split('\n') if line
                ] if deserialize else resp.body
                return result

            raise WALTailError(resp, request)

        return await self._execute(request, response_handler)



from __future__ import absolute_import, unicode_literals

import pytest

from aioarangodb.exceptions import (
    CollectionCreateError,
    CollectionListError,
    CollectionPropertiesError,
    DocumentInsertError,
    PermissionResetError,
    PermissionGetError,
    PermissionUpdateError,
    PermissionListError
)
from aioarangodb.tests.helpers import (
    assert_raises,
    extract,
    generate_col_name,
    generate_db_name,
    generate_string,
    generate_username
)
pytestmark = pytest.mark.asyncio


async def test_permission_management(client, sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    username = generate_username()
    password = generate_string()
    db_name = generate_db_name()
    col_name_1 = generate_col_name()
    col_name_2 = generate_col_name()

    await sys_db.create_database(
        name=db_name,
        users=[{
            'username': username,
            'password': password,
            'active': True
        }]
    )
    db = await client.db(db_name, username, password)
    assert isinstance(await sys_db.permissions(username), dict)

    # Test list permissions with bad database
    with assert_raises(PermissionListError) as err:
        await bad_db.permissions(username)
    assert err.value.error_code in {11, 1228}

    # Test get permission with bad database
    with assert_raises(PermissionGetError) as err:
        await bad_db.permission(username, db_name)
    assert err.value.error_code in {11, 1228}

    # The user should not have read and write permissions
    assert await sys_db.permission(username, db_name) == 'rw'
    assert await sys_db.permission(username, db_name, col_name_1) == 'rw'

    # Test update permission (database level) with bad database
    with assert_raises(PermissionUpdateError):
        await bad_db.update_permission(username, 'ro', db_name)
    assert await sys_db.permission(username, db_name) == 'rw'

    # Test update permission (database level) to read only and verify access
    assert await sys_db.update_permission(username, 'ro', db_name) is True
    assert await sys_db.permission(username, db_name) == 'ro'
    with assert_raises(CollectionCreateError) as err:
        await db.create_collection(col_name_2)
    assert err.value.http_code == 403
    assert col_name_1 not in await extract('name', await db.collections())
    assert col_name_2 not in await extract('name', await db.collections())

    # Test reset permission (database level) with bad database
    with assert_raises(PermissionResetError) as err:
        await bad_db.reset_permission(username, db_name)
    assert err.value.error_code in {11, 1228}
    assert await sys_db.permission(username, db_name) == 'ro'

    # Test reset permission (database level) and verify access
    assert await sys_db.reset_permission(username, db_name) is True
    assert await sys_db.permission(username, db_name) == 'none'
    with assert_raises(CollectionCreateError) as err:
        await db.create_collection(col_name_1)
    assert err.value.http_code == 401
    with assert_raises(CollectionListError) as err:
        await db.collections()
    assert err.value.http_code == 401

    # Test update permission (database level) and verify access
    assert await sys_db.update_permission(username, 'rw', db_name) is True
    assert await sys_db.permission(username, db_name, col_name_2) == 'rw'
    assert await db.create_collection(col_name_1) is not None
    assert await db.create_collection(col_name_2) is not None
    assert col_name_1 in await extract('name', await db.collections())
    assert col_name_2 in await extract('name', await db.collections())

    col_1 = await db.collection(col_name_1)
    col_2 = await db.collection(col_name_2)

    # Verify that user has read and write access to both collections
    assert isinstance(await col_1.properties(), dict)
    assert isinstance(await col_1.insert({}), dict)
    assert isinstance(await col_2.properties(), dict)
    assert isinstance(await col_2.insert({}), dict)

    # Test update permission (collection level) to read only and verify access
    assert await sys_db.update_permission(username, 'ro', db_name, col_name_1)
    assert await sys_db.permission(username, db_name, col_name_1) == 'ro'
    assert isinstance(await col_1.properties(), dict)
    with assert_raises(DocumentInsertError) as err:
        await col_1.insert({})
    assert err.value.http_code == 403
    assert isinstance(await col_2.properties(), dict)
    assert isinstance(await col_2.insert({}), dict)

    # Test update permission (collection level) to none and verify access
    assert await sys_db.update_permission(username, 'none', db_name, col_name_1)
    assert await sys_db.permission(username, db_name, col_name_1) == 'none'
    with assert_raises(CollectionPropertiesError) as err:
        await col_1.properties()
    assert err.value.http_code == 403
    with assert_raises(DocumentInsertError) as err:
        await col_1.insert({})
    assert err.value.http_code == 403
    assert isinstance(await col_2.properties(), dict)
    assert isinstance(await col_2.insert({}), dict)

    # Test reset permission (collection level)
    assert await sys_db.reset_permission(username, db_name, col_name_1) is True
    assert await sys_db.permission(username, db_name, col_name_1) == 'rw'
    assert isinstance(await col_1.properties(), dict)
    assert isinstance(await col_1.insert({}), dict)
    assert isinstance(await col_2.properties(), dict)
    assert isinstance(await col_2.insert({}), dict)



def verify_format(_, res):
    return res


def format_index(body):  # pragma: no cover
    """Format index data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        'id': body['id'].split('/', 1)[-1],
        'fields': body['fields']
    }
    if 'type' in body:
        result['type'] = body['type']
    if 'name' in body:
        result['name'] = body['name']
    if 'deduplicate' in body:
        result['deduplicate'] = body['deduplicate']
    if 'sparse' in body:
        result['sparse'] = body['sparse']
    if 'unique' in body:
        result['unique'] = body['unique']
    if 'minLength' in body:
        result['min_length'] = body['minLength']
    if 'geoJson' in body:
        result['geo_json'] = body['geoJson']
    if 'ignoreNull' in body:
        result['ignore_none'] = body['ignoreNull']
    if 'selectivityEstimate' in body:
        result['selectivity'] = body['selectivityEstimate']
    if 'isNewlyCreated' in body:
        result['new'] = body['isNewlyCreated']
    if 'expireAfter' in body:
        result['expiry_time'] = body['expireAfter']
    if 'inBackground' in body:
        result['in_background'] = body['inBackground']
    if 'bestIndexedLevel' in body:
        result['best_indexed_level'] = body['bestIndexedLevel']
    if 'worstIndexedLevel' in body:
        result['worst_indexed_level'] = body['worstIndexedLevel']
    if 'maxNumCoverCells' in body:
        result['max_num_cover_cells'] = body['maxNumCoverCells']

    return verify_format(body, result)


def format_key_options(body):  # pragma: no cover
    """Format collection key options data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if 'type' in body:
        result['key_generator'] = body['type']
    if 'increment' in body:
        result['key_increment'] = body['increment']
    if 'offset' in body:
        result['key_offset'] = body['offset']
    if 'allowUserKeys' in body:
        result['user_keys'] = body['allowUserKeys']
    if 'lastValue' in body:
        result['key_last_value'] = body['lastValue']

    return verify_format(body, result)


def format_database(body):  # pragma: no cover
    """Format databases info.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if 'id' in body:
        result['id'] = body['id']
    if 'name' in body:
        result['name'] = body['name']
    if 'path' in body:
        result['path'] = body['path']
    if 'system' in body:
        result['system'] = body['system']
    if 'isSystem' in body:
        result['system'] = body['isSystem']

    # Cluster only
    if 'sharding' in body:
        result['sharding'] = body['sharding']
    if 'replicationFactor' in body:
        result['replication_factor'] = body['replicationFactor']
    if 'writeConcern' in body:
        result['write_concern'] = body['writeConcern']

    return verify_format(body, result)


def format_collection(body):  # pragma: no cover
    """Format collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if 'id' in body:
        result['id'] = body['id']
    if 'objectId' in body:
        result['object_id'] = body['objectId']
    if 'name' in body:
        result['name'] = body['name']
    if 'isSystem' in body:
        result['system'] = body['isSystem']
    if 'isSmart' in body:
        result['smart'] = body['isSmart']
    if 'type' in body:
        result['type'] = body['type']
        result['edge'] = body['type'] == 3
    if 'waitForSync' in body:
        result['sync'] = body['waitForSync']

    if 'status' in body:
        result['status'] = body['status']
    if 'statusString' in body:
        result['status_string'] = body['statusString']
    if 'globallyUniqueId' in body:
        result['global_id'] = body['globallyUniqueId']
    if 'cacheEnabled' in body:
        result['cache'] = body['cacheEnabled']
    if 'replicationFactor' in body:
        result['replication_factor'] = body['replicationFactor']
    if 'minReplicationFactor' in body:
        result['min_replication_factor'] = body['minReplicationFactor']
    if 'writeConcern' in body:
        result['write_concern'] = body['writeConcern']

    # MMFiles only
    if 'doCompact' in body:
        result['compact'] = body['doCompact']
    if 'journalSize' in body:
        result['journal_size'] = body['journalSize']
    if 'isVolatile' in body:
        result['volatile'] = body['isVolatile']
    if 'indexBuckets' in body:
        result['index_bucket_count'] = body['indexBuckets']

    # Cluster only
    if 'shards' in body:
        result['shards'] = body['shards']
    if 'replicationFactor' in body:
        result['replication_factor'] = body['replicationFactor']
    if 'numberOfShards' in body:
        result['shard_count'] = body['numberOfShards']
    if 'shardKeys' in body:
        result['shard_fields'] = body['shardKeys']
    if 'distributeShardsLike' in body:
        result['shard_like'] = body['distributeShardsLike']
    if 'shardingStrategy' in body:
        result['sharding_strategy'] = body['shardingStrategy']
    if 'smartJoinAttribute' in body:
        result['smart_join_attribute'] = body['smartJoinAttribute']

    # Key Generator
    if 'keyOptions' in body:
        result['key_options'] = format_key_options(body['keyOptions'])

    # Replication only
    if 'cid' in body:
        result['cid'] = body['cid']
    if 'version' in body:
        result['version'] = body['version']
    if 'allowUserKeys' in body:
        result['user_keys'] = body['allowUserKeys']
    if 'planId' in body:
        result['plan_id'] = body['planId']
    if 'deleted' in body:
        result['deleted'] = body['deleted']

    # New in 3.7
    if 'syncByRevision' in body:
        result['sync_by_revision'] = body['syncByRevision']
    if 'tempObjectId' in body:
        result['temp_object_id'] = body['tempObjectId']
    if 'usesRevisionsAsDocumentIds' in body:
        result['rev_as_id'] = body['usesRevisionsAsDocumentIds']
    if 'isDisjoint' in body:
        result['disjoint'] = body['isDisjoint']
    if 'isSmartChild' in body:
        result['smart_child'] = body['isSmartChild']
    if 'minRevision' in body:
        result['min_revision'] = body['minRevision']
    if 'schema' in body:
        result['schema'] = body['schema']

    return verify_format(body, result)


def format_aql_cache(body):
    """Format AQL cache data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        'mode': body['mode'],
        'max_results': body['maxResults'],
        'max_results_size': body['maxResultsSize'],
        'max_entry_size': body['maxEntrySize'],
        'include_system': body['includeSystem']
    }
    return verify_format(body, result)


def format_wal_properties(body):  # pragma: no cover
    """Format WAL properties.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'allowOversizeEntries' in body:
        result['oversized_ops'] = body['allowOversizeEntries']
    if 'logfileSize' in body:
        result['log_size'] = body['logfileSize']
    if 'historicLogfiles' in body:
        result['historic_logs'] = body['historicLogfiles']
    if 'reserveLogfiles' in body:
        result['reserve_logs'] = body['reserveLogfiles']
    if 'syncInterval' in body:
        result['sync_interval'] = body['syncInterval']
    if 'throttleWait' in body:
        result['throttle_wait'] = body['throttleWait']
    if 'throttleWhenPending' in body:
        result['throttle_limit'] = body['throttleWhenPending']

    return verify_format(body, result)


def format_wal_transactions(body):  # pragma: no cover
    """Format WAL transactions.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'minLastCollected' in body:
        result['last_collected'] = body['minLastCollected']
    if 'minLastSealed' in body:
        result['last_sealed'] = body['minLastSealed']
    if 'runningTransactions' in body:
        result['count'] = body['runningTransactions']

    return verify_format(body, result)


def format_aql_query(body):  # pragma: no cover
    """Format AQL query data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {'id': body['id'], 'query': body['query']}
    if 'started' in body:
        result['started'] = body['started']
    if 'state' in body:
        result['state'] = body['state']
    if 'stream' in body:
        result['stream'] = body['stream']
    if 'bindVars' in body:
        result['bind_vars'] = body['bindVars']
    if 'runTime' in body:
        result['runtime'] = body['runTime']

    return verify_format(body, result)


def format_aql_tracking(body):  # pragma: no cover
    """Format AQL tracking data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'enabled' in body:
        result['enabled'] = body['enabled']
    if 'maxQueryStringLength' in body:
        result['max_query_string_length'] = body['maxQueryStringLength']
    if 'maxSlowQueries' in body:
        result['max_slow_queries'] = body['maxSlowQueries']
    if 'slowQueryThreshold' in body:
        result['slow_query_threshold'] = body['slowQueryThreshold']
    if 'slowStreamingQueryThreshold' in body:
        result['slow_streaming_query_threshold'] = \
            body['slowStreamingQueryThreshold']
    if 'trackBindVars' in body:
        result['track_bind_vars'] = body['trackBindVars']
    if 'trackSlowQueries' in body:
        result['track_slow_queries'] = body['trackSlowQueries']

    return verify_format(body, result)


def format_tick_values(body):  # pragma: no cover
    """Format tick data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}

    if 'tickMin' in body:
        result['tick_min'] = body['tickMin']
    if 'tickMax' in body:
        result['tick_max'] = body['tickMax']
    if 'tick' in body:
        result['tick'] = body['tick']
    if 'time' in body:
        result['time'] = body['time']
    if 'server' in body:
        result['server'] = format_server_info(body['server'])

    return verify_format(body, result)


def format_server_info(body):  # pragma: no cover
    """Format server data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    return {'version': body['version'], 'server_id': body['serverId']}


def format_replication_applier_config(body):  # pragma: no cover
    """Format replication applier configuration data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'endpoint' in body:
        result['endpoint'] = body['endpoint']
    if 'database' in body:
        result['database'] = body['database']
    if 'username' in body:
        result['username'] = body['username']
    if 'verbose' in body:
        result['verbose'] = body['verbose']
    if 'incremental' in body:
        result['incremental'] = body['incremental']
    if 'requestTimeout' in body:
        result['request_timeout'] = body['requestTimeout']
    if 'connectTimeout' in body:
        result['connect_timeout'] = body['connectTimeout']
    if 'ignoreErrors' in body:
        result['ignore_errors'] = body['ignoreErrors']
    if 'maxConnectRetries' in body:
        result['max_connect_retries'] = body['maxConnectRetries']
    if 'lockTimeoutRetries' in body:
        result['lock_timeout_retries'] = body['lockTimeoutRetries']
    if 'sslProtocol' in body:
        result['ssl_protocol'] = body['sslProtocol']
    if 'chunkSize' in body:
        result['chunk_size'] = body['chunkSize']
    if 'skipCreateDrop' in body:
        result['skip_create_drop'] = body['skipCreateDrop']
    if 'autoStart' in body:
        result['auto_start'] = body['autoStart']
    if 'adaptivePolling' in body:
        result['adaptive_polling'] = body['adaptivePolling']
    if 'autoResync' in body:
        result['auto_resync'] = body['autoResync']
    if 'autoResyncRetries' in body:
        result['auto_resync_retries'] = body['autoResyncRetries']
    if 'maxPacketSize' in body:
        result['max_packet_size'] = body['maxPacketSize']
    if 'includeSystem' in body:
        result['include_system'] = body['includeSystem']
    if 'includeFoxxQueues' in body:
        result['include_foxx_queues'] = body['includeFoxxQueues']
    if 'requireFromPresent' in body:
        result['require_from_present'] = body['requireFromPresent']
    if 'restrictType' in body:
        result['restrict_type'] = body['restrictType']
    if 'restrictCollections' in body:
        result['restrict_collections'] = body['restrictCollections']
    if 'connectionRetryWaitTime' in body:
        result['connection_retry_wait_time'] = body['connectionRetryWaitTime']
    if 'initialSyncMaxWaitTime' in body:
        result['initial_sync_max_wait_time'] = body['initialSyncMaxWaitTime']
    if 'idleMinWaitTime' in body:
        result['idle_min_wait_time'] = body['idleMinWaitTime']
    if 'idleMaxWaitTime' in body:
        result['idle_max_wait_time'] = body['idleMaxWaitTime']

    return verify_format(body, result)


def format_applier_progress(body):  # pragma: no cover
    """Format replication applier progress data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'time' in body:
        result['time'] = body['time']
    if 'message' in body:
        result['message'] = body['message']
    if 'failedConnects' in body:
        result['failed_connects'] = body['failedConnects']

    return verify_format(body, result)


def format_applier_error(body):  # pragma: no cover
    """Format replication applier error data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'errorNum' in body:
        result['error_num'] = body['errorNum']
    if 'errorMessage' in body:
        result['error_message'] = body['errorMessage']
    if 'time' in body:
        result['time'] = body['time']

    return verify_format(body, result)


def format_applier_state_details(body):  # pragma: no cover
    """Format replication applier state details.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'started' in body:
        result['started'] = body['started']
    if 'running' in body:
        result['running'] = body['running']
    if 'phase' in body:
        result['phase'] = body['phase']
    if 'time' in body:
        result['time'] = body['time']
    if 'safeResumeTick' in body:
        result['safe_resume_tick'] = body['safeResumeTick']
    if 'ticksBehind' in body:
        result['ticks_behind'] = body['ticksBehind']
    if 'lastAppliedContinuousTick' in body:
        result['last_applied_continuous_tick'] = \
            body['lastAppliedContinuousTick']
    if 'lastProcessedContinuousTick' in body:
        result['last_processed_continuous_tick'] = \
            body['lastProcessedContinuousTick']
    if 'lastAvailableContinuousTick' in body:
        result['last_available_continuous_tick'] = \
            body['lastAvailableContinuousTick']
    if 'progress' in body:
        result['progress'] = format_applier_progress(body['progress'])
    if 'totalRequests' in body:
        result['total_requests'] = body['totalRequests']
    if 'totalFailedConnects' in body:
        result['total_failed_connects'] = body['totalFailedConnects']
    if 'totalEvents' in body:
        result['total_events'] = body['totalEvents']
    if 'totalDocuments' in body:
        result['total_documents'] = body['totalDocuments']
    if 'totalRemovals' in body:
        result['total_removals'] = body['totalRemovals']
    if 'totalResyncs' in body:
        result['total_resyncs'] = body['totalResyncs']
    if 'totalOperationsExcluded' in body:
        result['total_operations_excluded'] = body['totalOperationsExcluded']
    if 'totalApplyTime' in body:
        result['total_apply_time'] = body['totalApplyTime']
    if 'averageApplyTime' in body:
        result['average_apply_time'] = body['averageApplyTime']
    if 'totalFetchTime' in body:
        result['total_fetch_time'] = body['totalFetchTime']
    if 'averageFetchTime' in body:
        result['average_fetch_time'] = body['averageFetchTime']
    if 'lastError' in body:
        result['last_error'] = format_applier_error(body['lastError'])

    return verify_format(body, result)


def format_replication_applier_state(body):  # pragma: no cover
    """Format replication applier state.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'endpoint' in body:
        result['endpoint'] = body['endpoint']
    if 'database' in body:
        result['database'] = body['database']
    if 'username' in body:
        result['username'] = body['username']
    if 'state' in body:
        result['state'] = format_applier_state_details(body['state'])
    if 'server' in body:
        result['server'] = format_server_info(body['server'])

    return verify_format(body, result)


def format_replication_state(body):  # pragma: no cover
    """Format replication state.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    if not isinstance(body, dict):
        return body

    result = {}
    if 'running' in body:
        result['running'] = body['running']
    if 'time' in body:
        result['time'] = body['time']
    if 'lastLogTick' in body:
        result['last_log_tick'] = body['lastLogTick']
    if 'totalEvents' in body:
        result['total_events'] = body['totalEvents']
    if 'lastUncommittedLogTick' in body:
        result['last_uncommitted_log_tick'] = body['lastUncommittedLogTick']

    return verify_format(body, result)


def format_replication_logger_state(body):  # pragma: no cover
    """Format replication collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'state' in body:
        result['state'] = format_replication_state(body['state'])
    if 'server' in body:
        result['server'] = format_server_info(body['server'])
    if 'clients' in body:
        result['clients'] = body['clients']

    return verify_format(body, result)


def format_replication_collection(body):  # pragma: no cover
    """Format replication collection data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'planVersion' in body:
        result['plan_version'] = body['planVersion']
    if 'isReady' in body:
        result['is_ready'] = body['isReady']
    if 'allInSync' in body:
        result['all_in_sync'] = body['allInSync']
    if 'indexes' in body:
        result['indexes'] = [format_index(index) for index in body['indexes']]
    if 'parameters' in body:
        result['parameters'] = format_collection(body['parameters'])

    return verify_format(body, result)


def format_replication_database(body):  # pragma: no cover
    """Format replication database data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {
        'id': body['id'],
        'name': body['name'],
        'collections': [
            format_replication_collection(col)
            for col in body['collections']
        ],
        'views': [format_view(view) for view in body['views']]
    }
    if 'properties' in body:
        result['properties'] = format_database(body['properties'])

    return verify_format(body, result)


def format_replication_inventory(body):  # pragma: no cover
    """Format replication inventory data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'tick' in body:
        result['tick'] = body['tick']
    if 'state' in body:
        result['state'] = format_replication_state(body['state'])

    if 'databases' in body:
        result['databases'] = {
            k: format_replication_database(v)
            for k, v in body['databases'].items()
        }
    if 'collections' in body:
        result['collections'] = [
            format_replication_collection(col)
            for col in body['collections']
        ]
    if 'views' in body:
        result['views'] = [format_view(view) for view in body['views']]
    if 'properties' in body:
        result['properties'] = format_database(body['properties'])

    return verify_format(body, result)


def format_replication_sync(body):  # pragma: no cover
    """Format replication sync result.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'collections' in body:
        result['collections'] = body['collections']
    if 'lastLogTick' in body:
        result['last_log_tick'] = body['lastLogTick']
    return verify_format(body, result)


def format_replication_header(headers):  # pragma: no cover
    """Format replication headers.

    :param headers: Request headers.
    :type headers: dict
    :return: Formatted body.
    :rtype: dict
    """
    headers = {k.lower(): v for k, v in headers.items()}
    result = {}

    if 'x-arango-replication-frompresent' in headers:
        result['from_present'] = \
            headers['x-arango-replication-frompresent'] == 'true'

    if 'x-arango-replication-lastincluded' in headers:
        result['last_included'] = \
            headers['x-arango-replication-lastincluded']

    if 'x-arango-replication-lastscanned' in headers:
        result['last_scanned'] = \
            headers['x-arango-replication-lastscanned']

    if 'x-arango-replication-lasttick' in headers:
        result['last_tick'] = \
            headers['x-arango-replication-lasttick']

    if 'x-arango-replication-active' in headers:
        result['active'] = \
            headers['x-arango-replication-active'] == 'true'

    if 'x-arango-replication-checkmore' in headers:
        result['check_more'] = \
            headers['x-arango-replication-checkmore'] == 'true'

    return result


def format_view_link(body):  # pragma: no cover
    """Format view link data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'analyzers' in body:
        result['analyzers'] = body['analyzers']
    if 'fields' in body:
        result['fields'] = body['fields']
    if 'includeAllFields' in body:
        result['include_all_fields'] = body['includeAllFields']
    if 'trackListPositions' in body:
        result['track_list_positions'] = body['trackListPositions']
    if 'storeValues' in body:
        result['store_values'] = body['storeValues']

    return verify_format(body, result)


def format_view_consolidation_policy(body):  # pragma: no cover
    """Format view consolidation policy data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'type' in body:
        result['type'] = body['type']
    if 'threshold' in body:
        result['threshold'] = body['threshold']
    if 'segmentsMin' in body:
        result['segments_min'] = body['segmentsMin']
    if 'segmentsMax' in body:
        result['segments_max'] = body['segmentsMax']
    if 'segmentsBytesMax' in body:
        result['segments_bytes_max'] = body['segmentsBytesMax']
    if 'segmentsBytesFloor' in body:
        result['segments_bytes_floor'] = body['segmentsBytesFloor']
    if 'minScore' in body:
        result['min_score'] = body['minScore']

    return verify_format(body, result)


def format_view(body):  # pragma: no cover
    """Format view data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = {}
    if 'globallyUniqueId' in body:
        result['global_id'] = body['globallyUniqueId']
    if 'id' in body:
        result['id'] = body['id']
    if 'name' in body:
        result['name'] = body['name']
    if 'type' in body:
        result['type'] = body['type']
    if 'cleanupIntervalStep' in body:
        result['cleanup_interval_step'] = body['cleanupIntervalStep']
    if 'commitIntervalMsec' in body:
        result['commit_interval_msec'] = body['commitIntervalMsec']
    if 'consolidationIntervalMsec' in body:
        result['consolidation_interval_msec'] = \
            body['consolidationIntervalMsec']
    if 'consolidationPolicy' in body:
        result['consolidation_policy'] = \
            format_view_consolidation_policy(body['consolidationPolicy'])
    if 'primarySort' in body:
        result['primary_sort'] = body['primarySort']
    if 'primarySortCompression' in body:
        result['primary_sort_compression'] = body['primarySortCompression']
    if 'storedValues' in body:
        result['stored_values'] = body['storedValues']
    if 'writebufferIdle' in body:
        result['writebuffer_idle'] = body['writebufferIdle']
    if 'writebufferActive' in body:
        result['writebuffer_active'] = body['writebufferActive']
    if 'writebufferSizeMax' in body:
        result['writebuffer_max_size'] = body['writebufferSizeMax']
    if 'links' in body:
        result['links'] = [format_view_link(link) for link in body['links']]

    return verify_format(body, result)


def format_vertex(body):  # pragma: no cover
    """Format vertex data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    vertex = body['vertex']
    if '_oldRev' in vertex:
        vertex['_old_rev'] = vertex.pop('_oldRev')

    if 'new' in body or 'old' in body:
        result = {'vertex': vertex}
        if 'new' in body:
            result['new'] = body['new']
        if 'old' in body:
            result['old'] = body['old']
        return result
    else:
        return vertex


def format_edge(body):  # pragma: no cover
    """Format edge data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    edge = body['edge']
    if '_oldRev' in edge:
        edge['_old_rev'] = edge.pop('_oldRev')

    if 'new' in body or 'old' in body:
        result = {'edge': edge}
        if 'new' in body:
            result['new'] = body['new']
        if 'old' in body:
            result['old'] = body['old']
        return result
    else:
        return edge


def format_tls(body):  # pragma: no cover
    """Format TLS data.

    :param body: Input body.
    :type body: dict
    :return: Formatted body.
    :rtype: dict
    """
    result = body
    return verify_format(body, result)



from __future__ import absolute_import, unicode_literals

__all__ = ['Response']


class Response(object):
    """HTTP response.

    :param method: HTTP method in lowercase (e.g. "post").
    :type method: str | unicode
    :param url: API URL.
    :type url: str | unicode
    :param headers: Response headers.
    :type headers: requests.structures.CaseInsensitiveDict | dict
    :param status_code: Response status code.
    :type status_code: int
    :param status_text: Response status text.
    :type status_text: str | unicode
    :param raw_body: Raw response body.
    :type raw_body: str | unicode

    :ivar method: HTTP method in lowercase (e.g. "post").
    :vartype method: str | unicode
    :ivar url: API URL.
    :vartype url: str | unicode
    :ivar headers: Response headers.
    :vartype headers: requests.structures.CaseInsensitiveDict | dict
    :ivar status_code: Response status code.
    :vartype status_code: int
    :ivar status_text: Response status text.
    :vartype status_text: str | unicode
    :ivar raw_body: Raw response body.
    :vartype raw_body: str | unicode
    :ivar body: JSON-deserialized response body.
    :vartype body: str | unicode | bool | int | list | dict
    :ivar error_code: Error code from ArangoDB server.
    :vartype error_code: int
    :ivar error_message: Error message from ArangoDB server.
    :vartype error_message: str | unicode
    :ivar is_success: True if response status code was 2XX.
    :vartype is_success: bool
    """

    __slots__ = (
        'method',
        'url',
        'headers',
        'status_code',
        'status_text',
        'body',
        'raw_body',
        'error_code',
        'error_message',
        'is_success',
    )

    def __init__(self,
                 method,
                 url,
                 headers,
                 status_code,
                 status_text,
                 raw_body):
        self.method = method.lower()
        self.url = url
        self.headers = headers
        self.status_code = status_code
        self.status_text = status_text
        self.raw_body = raw_body

        # Populated later
        self.body = None
        self.error_code = None
        self.error_message = None
        self.is_success = None



from __future__ import absolute_import, unicode_literals
import pytest
from six import string_types
from aioarangodb.connection import (
    BasicConnection,
    JWTConnection,
    JWTSuperuserConnection
)
from aioarangodb.exceptions import (
    JWTAuthError,
    # JWTSecretListError,
    # JWTSecretReloadError,
    ServerTLSError,
    ServerTLSReloadError,
    ServerVersionError
)
from aioarangodb.errno import FORBIDDEN, HTTP_UNAUTHORIZED
from aioarangodb.tests.helpers import assert_raises, generate_string, generate_jwt
pytestmark = pytest.mark.asyncio


async def test_auth_invalid_method(client, sys_db, db_name, username, password):
    with assert_raises(ValueError) as err:
        await client.db(
            name=db_name,
            username=username,
            password=password,
            verify=True,
            auth_method='bad_method'
        )
    assert 'invalid auth_method' in str(err.value)


async def test_auth_basic(client, sys_db, db_name, username, password):
    db = await client.db(
        name=db_name,
        username=username,
        password=password,
        verify=True,
        auth_method='basic'
    )
    assert isinstance(db.conn, BasicConnection)
    assert isinstance(await db.version(), string_types)
    assert isinstance(await db.properties(), dict)


async def test_auth_jwt(client, sys_db, db_name, username, password):
    db = await client.db(
        name=db_name,
        username=username,
        password=password,
        verify=True,
        auth_method='jwt'
    )
    assert isinstance(db.conn, JWTConnection)
    assert isinstance(await db.version(), string_types)
    assert isinstance(await db.properties(), dict)

    bad_password = generate_string()
    with assert_raises(JWTAuthError) as err:
        await client.db(db_name, username, bad_password, auth_method='jwt')
    assert err.value.error_code == HTTP_UNAUTHORIZED


async def test_auth_superuser_token(client, sys_db, db_name, root_password, secret):
    token = generate_jwt(secret)
    db = await client.db('_system', superuser_token=token)
    bad_db = await client.db('_system', superuser_token='bad_token')

    assert isinstance(db.conn, JWTSuperuserConnection)
    assert isinstance(await db.version(), string_types)
    assert isinstance(await db.properties(), dict)

    # # Test get JWT secrets
    # secrets = db.jwt_secrets()
    # assert 'active' in secrets
    # assert 'passive' in secrets
    #
    # # Test get JWT secrets with bad database
    # with assert_raises(JWTSecretListError) as err:
    #     bad_db.jwt_secrets()
    # assert err.value.error_code == FORBIDDEN
    #
    # # Test reload JWT secrets
    # secrets = db.reload_jwt_secrets()
    # assert 'active' in secrets
    # assert 'passive' in secrets
    #
    # # Test reload JWT secrets with bad database
    # with assert_raises(JWTSecretReloadError) as err:
    #     bad_db.reload_jwt_secrets()
    # assert err.value.error_code == FORBIDDEN

    # Test get TLS data
    result = await db.tls()
    assert isinstance(result, dict)

    # Test get TLS data with bad database
    with assert_raises(ServerTLSError) as err:
        await bad_db.tls()
    assert err.value.error_code == FORBIDDEN

    # Test reload TLS
    result = await db.reload_tls()
    assert isinstance(result, dict)

    # Test reload TLS with bad database
    with assert_raises(ServerTLSReloadError) as err:
        await bad_db.reload_tls()
    assert err.value.error_code == FORBIDDEN


async def test_auth_jwt_expiry(client, sys_db, db_name, root_password, secret):
    # Test automatic token refresh on expired token.
    db = await client.db('_system', 'root', root_password, auth_method='jwt')
    expired_token = generate_jwt(secret, exp=0)
    db.conn._token = expired_token
    db.conn._auth_header = 'bearer {}'.format(expired_token)
    assert isinstance(await db.version(), string_types)

    # Test correct error on token expiry.
    db = await client.db('_system', superuser_token=expired_token)
    with assert_raises(ServerVersionError) as err:
        await db.version()
    assert err.value.error_code == FORBIDDEN



from __future__ import absolute_import, unicode_literals

import time
import asyncio

import pytest
from six import string_types

from aioarangodb.database import AsyncDatabase
from aioarangodb.exceptions import (
    AsyncExecuteError,
    AsyncJobClearError,
    AsyncJobResultError,
    AsyncJobStatusError,
    AsyncJobListError,
    AsyncJobCancelError,
    AQLQueryExecuteError
)
from aioarangodb.job import AsyncJob
from aioarangodb.tests.helpers import extract
pytestmark = pytest.mark.asyncio


async def wait_on_job(job):
    """Block until the async job is done."""
    while await job.status() != 'done':
        await asyncio.sleep(.05)  # pragma: no cover
    return job


async def wait_on_jobs(db):
    """Block until all async jobs are finished."""
    while len(await db.async_jobs('pending')) > 0:
        await asyncio.sleep(.05)  # pragma: no cover


async def test_async_wrapper_attributes(db, col, username):
    async_db = db.begin_async_execution()
    assert isinstance(async_db, AsyncDatabase)
    assert async_db.username == username
    assert async_db.context == 'async'
    assert async_db.db_name == db.name
    assert async_db.name == db.name
    assert repr(async_db) == '<AsyncDatabase {}>'.format(db.name)

    async_col = async_db.collection(col.name)
    assert async_col.username == username
    assert async_col.context == 'async'
    assert async_col.db_name == db.name
    assert async_col.name == col.name

    async_aql = async_db.aql
    assert async_aql.username == username
    assert async_aql.context == 'async'
    assert async_aql.db_name == db.name

    job = await async_aql.execute('INVALID QUERY')
    assert isinstance(job, AsyncJob)
    assert isinstance(job.id, string_types)
    assert repr(job) == '<AsyncJob {}>'.format(job.id)


async def test_async_execute_without_result(db, col, docs):
    # Insert test documents asynchronously with return_result set to False
    async_col = db.begin_async_execution(return_result=False).collection(
        col.name)

    # Ensure that no jobs were returned
    assert await async_col.insert(docs[0]) is None
    assert await async_col.insert(docs[1]) is None
    assert await async_col.insert(docs[2]) is None

    # Ensure that the operations went through
    await wait_on_jobs(db)
    assert await extract('_key', await col.all()) == ['1', '2', '3']


async def test_async_execute_error_in_result(db, col, docs):
    await db.collection(col.name).import_bulk(docs)
    async_db = db.begin_async_execution(return_result=True)

    # Test async execution of a bad AQL query
    job = await wait_on_job(await async_db.aql.execute('INVALID QUERY'))
    with pytest.raises(AQLQueryExecuteError) as err:
        await job.result()
    assert err.value.error_code == 1501


async def test_async_get_job_status(db, bad_db):
    async_db = db.begin_async_execution(return_result=True)

    # Test get status of a pending job
    job = await async_db.aql.execute('RETURN SLEEP(0.1)', count=True)
    assert await job.status() == 'pending'

    # Test get status of a finished job
    assert await (await wait_on_job(job)).status() == 'done'
    assert (await job.result()).count() == 1

    # Test get status of a missing job
    with pytest.raises(AsyncJobStatusError) as err:
        await job.status()
    assert err.value.error_code == 404

    # Test get status from invalid job
    bad_job = await wait_on_job(await async_db.aql.execute('INVALID QUERY'))
    bad_job._conn = bad_db._conn
    with pytest.raises(AsyncJobStatusError) as err:
        await bad_job.status()
    assert err.value.error_code in {11, 1228}


async def test_async_get_job_result(db, bad_db):
    async_db = db.begin_async_execution(return_result=True)

    # Test get result from a pending job
    job = await async_db.aql.execute('RETURN SLEEP(0.1)', count=True)
    with pytest.raises(AsyncJobResultError) as err:
        await job.result()
    assert err.value.http_code == 204
    assert '{} not done'.format(job.id) in str(err.value)

    # Test get result from a finished job
    assert (await (await wait_on_job(job)).result()).count() == 1

    # Test get result from a cleared job
    with pytest.raises(AsyncJobResultError) as err:
        await job.result()
    assert err.value.error_code == 404

    # Test get result from an invalid job
    bad_job = await async_db.aql.execute('INVALID QUERY')
    bad_job._conn = bad_db._conn
    with pytest.raises(AsyncJobResultError) as err:
        await bad_job.result()
    assert err.value.error_code in {11, 1228}


async def test_async_cancel_job(db, bad_db):
    async_db = db.begin_async_execution(return_result=True)

    # Start a long running request to ensure that job can be cancelled
    job = await async_db.aql.execute('RETURN SLEEP(5)')

    # Test cancel a pending job
    assert await job.cancel() is True

    # Test cancel a missing job
    job._id = 'invalid_id'
    with pytest.raises(AsyncJobCancelError) as err:
        await job.cancel(ignore_missing=False)
    assert err.value.error_code == 404
    assert await job.cancel(ignore_missing=True) is False

    # Test cancel an invalid job
    job = await async_db.aql.execute('RETURN SLEEP(5)')
    job._conn = bad_db._conn
    with pytest.raises(AsyncJobCancelError) as err:
        await job.cancel()
    assert err.value.error_code in {11, 1228}


async def test_async_clear_job(db, bad_db):
    async_db = db.begin_async_execution(return_result=True)

    job = await async_db.aql.execute('RETURN 1')

    # Test clear finished job
    assert await job.clear(ignore_missing=True) is True

    # Test clear missing job
    with pytest.raises(AsyncJobClearError) as err:
        await job.clear(ignore_missing=False)
    assert err.value.error_code == 404
    assert await job.clear(ignore_missing=True) is False

    # Test clear with an invalid job
    job._conn = bad_db._conn
    with pytest.raises(AsyncJobClearError) as err:
        await job.clear()
    assert err.value.error_code in {11, 1228}


async def test_async_execute_errors(bad_db):
    bad_async_db = bad_db.begin_async_execution(return_result=False)
    with pytest.raises(AsyncExecuteError) as err:
        await bad_async_db.aql.execute('RETURN 1')
    assert err.value.error_code in {11, 1228}

    bad_async_db = bad_db.begin_async_execution(return_result=True)
    with pytest.raises(AsyncExecuteError) as err:
        await bad_async_db.aql.execute('RETURN 1')
    assert err.value.error_code in {11, 1228}


async def test_async_clear_jobs(db, bad_db, col, docs):
    async_db = db.begin_async_execution(return_result=True)
    async_col = async_db.collection(col.name)

    job1 = await wait_on_job(await async_col.insert(docs[0]))
    job2 = await wait_on_job(await async_col.insert(docs[1]))
    job3 = await wait_on_job(await async_col.insert(docs[2]))

    # Test clear all async jobs
    assert await db.clear_async_jobs() is True
    for job in [job1, job2, job3]:
        with pytest.raises(AsyncJobStatusError) as err:
            await job.status()
        assert err.value.error_code == 404

    # Set up test documents again
    job1 = await wait_on_job(await async_col.insert(docs[0]))
    job2 = await wait_on_job(await async_col.insert(docs[1]))
    job3 = await wait_on_job(await async_col.insert(docs[2]))

    # Test clear jobs that have expired
    past = int(time.time()) - 1000000
    assert await db.clear_async_jobs(threshold=past) is True
    for job in [job1, job2, job3]:
        assert await job.status() == 'done'

    # Test clear jobs that have not expired yet
    future = int(time.time()) + 1000000
    assert await db.clear_async_jobs(threshold=future) is True
    for job in [job1, job2, job3]:
        with pytest.raises(AsyncJobStatusError) as err:
            await job.status()
        assert err.value.error_code == 404

    # Test clear job with bad database
    with pytest.raises(AsyncJobClearError) as err:
        await bad_db.clear_async_jobs()
    assert err.value.error_code in {11, 1228}


async def test_async_list_jobs(db, col, docs):
    async_db = db.begin_async_execution(return_result=True)
    async_col = async_db.collection(col.name)

    job1 = await wait_on_job(await async_col.insert(docs[0]))
    job2 = await wait_on_job(await async_col.insert(docs[1]))
    job3 = await wait_on_job(await async_col.insert(docs[2]))

    # Test list async jobs that are done
    job_ids = await db.async_jobs(status='done')
    assert job1.id in job_ids
    assert job2.id in job_ids
    assert job3.id in job_ids

    # Test list async jobs that are pending
    job4 = await async_db.aql.execute('RETURN SLEEP(0.3)')
    assert await db.async_jobs(status='pending') == [job4.id]
    await wait_on_job(job4)  # Make sure the job is done

    # Test list async jobs with invalid status
    with pytest.raises(AsyncJobListError) as err:
        await db.async_jobs(status='bad_status')
    assert err.value.error_code == 400

    # Test list jobs with count
    job_ids = await db.async_jobs(status='done', count=1)
    assert len(job_ids) == 1
    assert job_ids[0] in [job1.id, job2.id, job3.id, job4.id]



from __future__ import absolute_import, unicode_literals

import pytest

from aioarangodb.errno import (
    FORBIDDEN,
    HTTP_UNAUTHORIZED,
    DATABASE_NOT_FOUND
)
from aioarangodb.exceptions import (
    WALConfigureError,
    WALFlushError,
    WALPropertiesError,
    WALTransactionListError,
    WALTickRangesError,
    WALLastTickError,
    WALTailError
)
from aioarangodb.tests.helpers import assert_raises
pytestmark = pytest.mark.asyncio


async def test_wal_misc_methods(sys_db, bad_db):
    try:
        await sys_db.wal.properties()
    except WALPropertiesError as err:
        if err.http_code == 501:
            pytest.skip('WAL not implemented')

    # Test get properties
    properties = await sys_db.wal.properties()
    assert 'oversized_ops' in properties
    assert 'log_size' in properties
    assert 'historic_logs' in properties
    assert 'reserve_logs' in properties
    assert 'throttle_wait' in properties
    assert 'throttle_limit' in properties

    # Test get properties with bad database
    with assert_raises(WALPropertiesError) as err:
        await bad_db.wal.properties()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test configure properties
    await sys_db.wal.configure(
        historic_logs=15,
        oversized_ops=False,
        log_size=30000000,
        reserve_logs=5,
        throttle_limit=0,
        throttle_wait=16000
    )
    properties = await sys_db.wal.properties()
    assert properties['historic_logs'] == 15
    assert properties['oversized_ops'] is False
    assert properties['log_size'] == 30000000
    assert properties['reserve_logs'] == 5
    assert properties['throttle_limit'] == 0
    assert properties['throttle_wait'] == 16000

    # Test configure properties with bad database
    with assert_raises(WALConfigureError) as err:
        await bad_db.wal.configure(log_size=2000000)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test get transactions
    result = await sys_db.wal.transactions()
    assert 'count' in result
    assert 'last_collected' in result

    # Test get transactions with bad database
    with assert_raises(WALTransactionListError) as err:
        await bad_db.wal.transactions()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}

    # Test flush
    result = await sys_db.wal.flush(garbage_collect=False, sync=False)
    assert isinstance(result, bool)

    # Test flush with bad database
    with assert_raises(WALFlushError) as err:
        await bad_db.wal.flush(garbage_collect=False, sync=False)
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_wal_tick_ranges(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await sys_db.wal.tick_ranges()
    assert 'server' in result
    assert 'time' in result
    assert 'tick_min' in result
    assert 'tick_max' in result

    # Test tick_ranges with bad database
    with assert_raises(WALTickRangesError) as err:
        await bad_db.wal.tick_ranges()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_wal_last_tick(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await sys_db.wal.last_tick()
    assert 'time' in result
    assert 'tick' in result
    assert 'server' in result

    # Test last_tick with bad database
    with assert_raises(WALLastTickError) as err:
        await bad_db.wal.last_tick()
    assert err.value.error_code in {FORBIDDEN, DATABASE_NOT_FOUND}


async def test_wal_tail(sys_db, bad_db, cluster):
    if cluster:
        pytest.skip('Not tested in a cluster setup')

    result = await sys_db.wal.tail(
        lower=0,
        upper=1000000,
        last_scanned=0,
        all_databases=True,
        chunk_size=1000000,
        syncer_id=None,
        server_id=None,
        client_info='test',
        barrier_id=None
    )
    assert 'content' in result
    assert 'last_tick' in result
    assert 'last_scanned' in result
    assert 'last_included' in result
    assert isinstance(result['check_more'], bool)
    assert isinstance(result['from_present'], bool)

    # Test tick_ranges with bad database
    with assert_raises(WALTailError) as err:
        await bad_db.wal.tail()
    assert err.value.http_code == HTTP_UNAUTHORIZED
